@@ -15,12 +15,12 @@ use tokio::{
 };
 use tokio_stream::*;
 
-use crate::header::BLOCK_SIZE;
 use crate::{
     entry::{EntryFields, EntryIo},
     error::TarError,
     other, Entry, GnuExtSparseHeader, GnuSparseHeader, Header,
 };
+use crate::{header::BLOCK_SIZE, pax::pax_extensions};
 
 /// A top-level representation of an archive file.
 ///
@@ -44,6 +44,7 @@ pub struct ArchiveInner<R> {
     unpack_xattrs: bool,
     preserve_permissions: bool,
     preserve_mtime: bool,
+    allow_external_symlinks: bool,
     overwrite: bool,
     ignore_zeros: bool,
     obj: Mutex<R>,
@@ -55,6 +56,7 @@ pub struct ArchiveBuilder<R: Read + Unpin> {
     unpack_xattrs: bool,
     preserve_permissions: bool,
     preserve_mtime: bool,
+    allow_external_symlinks: bool,
     overwrite: bool,
     ignore_zeros: bool,
 }
@@ -66,6 +68,7 @@ impl<R: Read + Unpin> ArchiveBuilder<R> {
             unpack_xattrs: false,
             preserve_permissions: false,
             preserve_mtime: true,
+            allow_external_symlinks: true,
             overwrite: true,
             ignore_zeros: false,
             obj,
@@ -118,12 +121,23 @@ impl<R: Read + Unpin> ArchiveBuilder<R> {
         self
     }
 
+    /// Indicate whether to deny symlinks that point outside the destination
+    /// directory when unpacking this entry. (Writing to locations outside the
+    /// destination directory is _always_ forbidden.)
+    ///
+    /// This flag is enabled by default.
+    pub fn set_allow_external_symlinks(mut self, allow_external_symlinks: bool) -> Self {
+        self.allow_external_symlinks = allow_external_symlinks;
+        self
+    }
+
     /// Construct the archive, ready to accept inputs.
     pub fn build(self) -> Archive<R> {
         let Self {
             unpack_xattrs,
             preserve_permissions,
             preserve_mtime,
+            allow_external_symlinks,
             overwrite,
             ignore_zeros,
             obj,
@@ -134,6 +148,7 @@ impl<R: Read + Unpin> ArchiveBuilder<R> {
                 unpack_xattrs,
                 preserve_permissions,
                 preserve_mtime,
+                allow_external_symlinks,
                 overwrite,
                 ignore_zeros,
                 obj: Mutex::new(obj),
@@ -151,6 +166,7 @@ impl<R: Read + Unpin> Archive<R> {
                 unpack_xattrs: false,
                 preserve_permissions: false,
                 preserve_mtime: true,
+                allow_external_symlinks: true,
                 overwrite: true,
                 ignore_zeros: false,
                 obj: Mutex::new(obj),
@@ -186,7 +202,7 @@ impl<R: Read + Unpin> Archive<R> {
         Ok(Entries {
             archive: self.clone(),
             pending: None,
-            current: (0, None, 0, None),
+            current: (0, None, 0, None, None),
             gnu_longlink: (false, None),
             gnu_longname: (false, None),
             pax_extensions: (false, None),
@@ -289,7 +305,13 @@ impl<R: Read + Unpin> Archive<R> {
 /// Stream of `Entry`s.
 pub struct Entries<R: Read + Unpin> {
     archive: Archive<R>,
-    current: (u64, Option<Header>, usize, Option<GnuExtSparseHeader>),
+    current: (
+        u64,
+        Option<Header>,
+        usize,
+        Option<GnuExtSparseHeader>,
+        Option<Vec<u8>>,
+    ),
     /// The [`Entry`] that is currently being processed.
     pending: Option<Entry<Archive<R>>>,
     /// GNU long name extension.
@@ -341,13 +363,15 @@ impl<R: Read + Unpin> Stream for Entries<R> {
             let entry = if let Some(entry) = self.pending.take() {
                 entry
             } else {
-                let (next, current_header, current_header_pos, _) = &mut self.current;
+                let (next, current_header, current_header_pos, _, pax_extensions) =
+                    &mut self.current;
                 ready_opt_err!(poll_next_raw(
                     archive,
                     next,
                     current_header,
                     current_header_pos,
-                    cx
+                    cx,
+                    pax_extensions.as_deref(),
                 ))
             };
 
@@ -429,6 +453,7 @@ impl<R: Read + Unpin> Stream for Entries<R> {
                 }
 
                 self.pax_extensions.0 = true;
+                self.current.4 = self.pax_extensions.1.clone();
                 continue;
             }
 
@@ -447,7 +472,7 @@ impl<R: Read + Unpin> Stream for Entries<R> {
             }
 
             let archive = self.archive.clone();
-            let (next, _, current_pos, current_ext) = &mut self.current;
+            let (next, _, current_pos, current_ext, _pax_extensions) = &mut self.current;
 
             ready_err!(poll_parse_sparse_header(
                 archive,
@@ -455,7 +480,7 @@ impl<R: Read + Unpin> Stream for Entries<R> {
                 current_ext,
                 current_pos,
                 &mut fields,
-                cx
+                cx,
             ));
 
             return Poll::Ready(Some(Ok(fields.into_entry())));
@@ -475,7 +500,7 @@ impl<R: Read + Unpin> Stream for RawEntries<R> {
     fn poll_next(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
         let archive = self.archive.clone();
         let (next, current_header, current_header_pos) = &mut self.current;
-        poll_next_raw(archive, next, current_header, current_header_pos, cx)
+        poll_next_raw(archive, next, current_header, current_header_pos, cx, None)
     }
 }
 
@@ -485,6 +510,7 @@ fn poll_next_raw<R: Read + Unpin>(
     current_header: &mut Option<Header>,
     current_header_pos: &mut usize,
     cx: &mut Context<'_>,
+    pax_extensions_data: Option<&[u8]>,
 ) -> Poll<Option<io::Result<Entry<Archive<R>>>>> {
     let mut header_pos = *next;
 
@@ -545,13 +571,66 @@ fn poll_next_raw<R: Read + Unpin>(
     }
 
     let file_pos = *next;
-    let size = header.entry_size()?;
+
+    let mut header = current_header.take().unwrap();
+
+    // note when pax extensions are available, the size from the header will be ignored
+    let mut size = header.entry_size()?;
+
+    // the size above will be overriden by the pax data if it has a size field.
+    // same for uid and gid, which will be overridden in the header itself.
+    if let Some(pax) = pax_extensions_data.map(pax_extensions) {
+        for extension in pax {
+            let extension = extension?;
+
+            // ignore keys that aren't parsable as a string at this stage.
+            // that isn't relevant to the size/uid/gid processing.
+            let Ok(key) = extension.key() else {
+                continue;
+            };
+
+            match key {
+                "size" => {
+                    let size_str = extension
+                        .value()
+                        .map_err(|_e| other("failed to parse pax size as string"))?;
+                    size = size_str
+                        .parse::<u64>()
+                        .map_err(|_e| other("failed to parse pax size"))?;
+                }
+
+                "uid" => {
+                    let uid_str = extension
+                        .value()
+                        .map_err(|_e| other("failed to parse pax uid as string"))?;
+                    header.set_uid(
+                        uid_str
+                            .parse::<u64>()
+                            .map_err(|_e| other("failed to parse pax uid"))?,
+                    );
+                }
+
+                "gid" => {
+                    let gid_str = extension
+                        .value()
+                        .map_err(|_e| other("failed to parse pax gid as string"))?;
+                    header.set_gid(
+                        gid_str
+                            .parse::<u64>()
+                            .map_err(|_e| other("failed to parse pax gid"))?,
+                    );
+                }
+
+                _ => {
+                    continue;
+                }
+            }
+        }
+    }
 
     let mut data = VecDeque::with_capacity(1);
     data.push_back(EntryIo::Data(archive.clone().take(size)));
 
-    let header = current_header.take().unwrap();
-
     let ret = EntryFields {
         size,
         header_pos,
@@ -565,6 +644,7 @@ fn poll_next_raw<R: Read + Unpin>(
         preserve_permissions: archive.inner.preserve_permissions,
         preserve_mtime: archive.inner.preserve_mtime,
         overwrite: archive.inner.overwrite,
+        allow_external_symlinks: archive.inner.allow_external_symlinks,
         read_state: None,
     };
 
